// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Entity.hpp"
#include "scene/Scene.hpp"
#include "assert/assert.hpp"
#include "util/Easing.hpp"

using namespace Dawn;

const glm::vec3 EntityTilePosition::toWorldSpace() {
  return glm::vec3(
    this->x,
    this->y,
    this->z
  );
}

void Entity::onInit() {
  auto world = this->world.lock();
  assertNotNull(world, "World is no longer active?");

  listeners.push_back(getScene()->onUnpausedUpdate.listen([this](float delta) {
    if(this->stepTime <= 0.0f) return;
    this->stepTime -= delta * this->stepSpeed;

    glm::vec3 nextPosition = this->tilePosition.toWorldSpace();
    
    if(this->stepTime <= 0) {
      this->getItem()->setLocalPosition(nextPosition);
      this->stepTime = 0.0f;
      this->eventMove.emit();
      this->eventStepEnd.emit();
      return;
    }

    float_t t = Easing::easeInOutSine(1.0f - this->stepTime);
    glm::vec3 lastPosition = this->lastTilePosition.toWorldSpace();
    glm::vec3 newLocal = glm::mix(lastPosition, nextPosition, t);
    this->getItem()->setLocalPosition(newLocal);
    this->eventMove.emit();
  }));
}

void Entity::onDispose() {
  
}

void Entity::move(
  const enum EntityDirection direction,
  const float_t stepSpeed
) {
  struct EntityTilePosition newPosition = this->tilePosition;

  switch(direction) {
    case EntityDirection::Up:
      newPosition.z--;
      break;
    case EntityDirection::Down:
      newPosition.z++;
      break;
    case EntityDirection::Left:
      newPosition.x--;
      break;
    case EntityDirection::Right:
      newPosition.x++;
      break;
    default:
      assertUnreachable("Invalid direction: %d", direction);
  }

  // Get the tile at the destination, check for height if we are on stairs, etc.
  // If the tile is not walkable, return early.

  // Move the entity to the new tile.
  this->lastTilePosition = this->tilePosition;
  this->tilePosition = newPosition;
  this->direction = direction;
  this->stepTime = 1.0f;
  this->stepSpeed = stepSpeed;

  this->eventStepStart.emit();
}

void Entity::setPosition(struct EntityTilePosition pos) {
  this->stepTime = 0.0f;
  this->tilePosition = pos;
  this->lastTilePosition = pos;
  this->getItem()->setLocalPosition(pos.toWorldSpace());
}

bool_t Entity::isMoving() {
  return this->stepTime > 0.0f;
}